#include <ssslib.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <sys/time.h>
#include <unistd.h>
#include <time.h>

#ifdef __APPLE__
typedef int socklen_t;
#endif

void send_event(SSS sss,int i,char *handle){
  int sendbuf_len;
  char *sendbuf;
  char *msg;
  int rc;

  time_t tv;

  tv = time(NULL);
  
  sendbuf_len = snprintf(NULL,0,"<add-subscription><subscription component='e-pingf' msg='ping' notify='e-pingf' data='%i'/></add-subscription>",i);

  sendbuf = malloc(sendbuf_len+1);

  sprintf(sendbuf,"<add-subscription><subscription component='e-pingf' msg='ping' notify='e-pingf' data='%i'/></add-subscription>",i);

  rc = SSS_SendMessage(sss,handle,sendbuf);
  if (rc == -1){
    printf("Error in SendMessage\n");
    exit(1);
  }

  free(sendbuf);

  msg = SSS_RecvMessage(sss,handle);

  free(msg);

  sendbuf_len = snprintf(NULL,0,"<add-event><event component='e-pingf' msg='ping' data='%i' time='%li'/></add-event>",i,tv) + 1;
  sendbuf = malloc(sendbuf_len);
  sprintf(sendbuf,"<add-event><event component='e-pingf' msg='ping' data='%i' time='%li'/></add-event>",i,tv);
  SSS_SendMessage(sss,handle,sendbuf);

  free(sendbuf);

}

int main(int argc, char **argv){
  SSS sss;
  char *clihandle;
  char *servhandle;
  char bufmsg[10000];
  char *msg;
  int i,j;
  int total;
  int rc;
  struct timeval short_tv;
  struct timeval long_tv;

  fd_set fds;

  int servsock;
  int clisock;

  int yes;
  
  struct sockaddr_in servaddr;
  struct sockaddr_in cliaddr;

  socklen_t length = sizeof(struct sockaddr);

  if (argc < 2 || argc > 3){
    printf("Usage: e-pingf <number of times> [<debug>]\n");
    exit(1);
  }

  total = atoi(argv[1]);

  /* init sockets */
  
  servsock = socket(PF_INET,SOCK_STREAM,IPPROTO_TCP);
  if (servsock == -1){
    perror("socket");
    exit(1);
  }

  servaddr.sin_family = AF_INET;
  servaddr.sin_port = htons(9000);
  servaddr.sin_addr.s_addr = INADDR_ANY;
  yes = 1;

  setsockopt(servsock,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int));

  rc = bind(servsock,(struct sockaddr *)&servaddr,sizeof(struct sockaddr));
  if (rc == -1){
    perror("bind");
    exit(1);
  }
  
  rc = listen(servsock,500);
  if (rc == -1){
    perror("listen");
    exit(1);
  }
  
  short_tv.tv_sec = 0;
  short_tv.tv_usec = 100000;

  long_tv.tv_sec = 1;
  long_tv.tv_usec = 0;

  if (argc == 3)
    sss = SSS_Init(NULL,atoi(argv[2]));
  else
    sss = SSS_Init(NULL,0);
  if (!sss){
    printf("Error in sss init\n");
    exit(1);
  }

  rc = SSS_SDRegister(sss,"e-pingf","localhost",9000,"basic","1.0","1");
  if (rc == -1){
    printf("Error in SDRegister\n");
    exit(1);
  }

  clihandle = SSS_ClientInit(sss,"event-manager");
  if (!clihandle){
    SSS_perror(sss,"ClientInit");
    exit(1);
  }
  
  FD_ZERO(&fds);

  FD_SET(servsock,&fds);

  i = 0;
  j = 0;

  while (1) {

    rc = select(servsock+1,&fds,NULL,NULL,(j == total) ? &long_tv : &short_tv);
    if (rc == -1){
      perror("select");
      exit(1);
    }

    if ((!rc) && (j == total)){
      break;
    }
    
    if (FD_ISSET(servsock,&fds)){
      clisock = accept(servsock,(struct sockaddr *)&cliaddr,&length);
      if (clisock == -1){
	perror("accept");
	exit(1);
      }
      servhandle = SSS_ServerInit(sss,clisock);
      if (!servhandle){
	SSS_perror(sss,"ServerInit");
	exit(1);
      }

      msg = SSS_RecvMessage(sss,servhandle);

      if (!msg){
	rc = close(clisock);
	if (rc == -1)
	  perror("close");
      }
      else{
	i--;
	rc = SSS_SendMessage(sss,servhandle,"<event-ok/>");
	if (rc == -1)
	  perror("send");
      }
      rc = SSS_ServerClose(sss,servhandle);
      if (rc == -1){
	printf("Error in ServerClose\n");
	exit(1);
      }
    }
    if (j < total){
      send_event(sss,j++,clihandle);
      i++;
    }
    FD_ZERO(&fds);
    FD_SET(servsock,&fds);
    short_tv.tv_sec = 0;
    short_tv.tv_usec = 100000;
    long_tv.tv_sec = 1;
    long_tv.tv_usec = 0;
  }

  printf("Finished run, sent %i, received %i back\n",total,total-i);

  SSS_ClientClose(sss,clihandle);

  SSS_SDRemove(sss,"e-pingf","localhost",9000,"basic");

  SSS_Close(sss);

  return 0;
}
