/*
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 1999-2001 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xerces" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache\@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation, and was
 * originally based on software copyright (c) 1999, International
 * Business Machines, Inc., http://www.ibm.com .  For more information
 * on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

/*
 * $Id: DOMCount.cpp,v 1.22 2002/02/01 22:35:01 peiyongz Exp $
 */

// ---------------------------------------------------------------------------
//  Includes
// ---------------------------------------------------------------------------
#include <xercesc/util/PlatformUtils.hpp>
#include <xercesc/sax/SAXException.hpp>
#include <xercesc/sax/SAXParseException.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/dom/DOMException.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
#include <string.h>
#include <stdlib.h>
#include <fstream.h>
#include <stdio.h>
#include "validate.hpp"

extern "C" {

  char *schema_dir = SCHEMA_DIR;

  char *err_string;

  char *validate_xml(char *xml_buf,char *schema,char *errbuf,int errlen)
  {

    int schemaloc_size;

    schemaloc_size = snprintf(NULL,0,"%s/%s",SCHEMA_DIR,schema)+1;
    
    char *realSchemaLoc;
    if (!(schema[0] == '/')){
      realSchemaLoc = (char *)malloc(schemaloc_size);
      sprintf(realSchemaLoc,"%s/%s",SCHEMA_DIR,schema);
    }
    else {
      realSchemaLoc = strdup(schema);
    }

    // Initialize the XML4C system
    try
      {
        XMLPlatformUtils::Initialize();
      }

    catch (const XMLException& toCatch)
      {
	char *msg = (char *)toCatch.getMessage();
	strncpy(errbuf,msg,errlen);
	free(realSchemaLoc);
	return err_string;
      }

    err_string = strdup("");

    const char*              xmlFile = 0;
    XercesDOMParser::ValSchemes    valScheme = XercesDOMParser::Val_Auto;
    bool                     doNamespaces       = true;
    bool                     doSchema           = true;
    bool                     schemaFullChecking = true;
    bool                     doList = false;
    bool                     errorOccurred = false;

    // Instantiate the DOM parser.
    XercesDOMParser* parser = new XercesDOMParser;
    parser->setValidationScheme(valScheme);
    parser->setDoNamespaces(doNamespaces);
    parser->setDoSchema(doSchema);
    parser->setValidationSchemaFullChecking(schemaFullChecking);
    parser->setExternalNoNamespaceSchemaLocation(realSchemaLoc); 
    
    // And create our error handler and install it
    DOMCountErrorHandler errorHandler;
    parser->setErrorHandler(&errorHandler);

    bool more = true;
    ifstream fin;

    MemBufInputSource *is = new MemBufInputSource((XMLByte *)xml_buf,strlen(xml_buf),"Xerces");
    
    while (more)
      {
	more = false;

        //reset error count first
        errorHandler.resetErrors();

        try
	  {
            parser->parse(*is);
	  }

        catch (const XMLException& toCatch)
	  {
	    char *msg = (char *)toCatch.getMessage();
	    strncpy(errbuf,msg,errlen);
            errorOccurred = true;
            continue;
	  }
        catch (const DOMException& toCatch)
	  {
	    strncpy(errbuf,"DOMException",errlen);
            errorOccurred = true;
            continue;
	  }
        catch (...)
	  {
	    strncpy(errbuf,"Unexpected internal error",errlen);
            errorOccurred = true;
            continue;
	  }

        //
        //  Extract the DOM tree, get the list of all the elements and report the
        //  length as the count of elements.
        //
        if (errorHandler.getSawErrors())
	  {
	    strncpy(errbuf,"Unexpected internal error",errlen);
            errorOccurred = true;
	  }
      }

    if (doList)
      fin.close();

    //
    //  Delete the parser itself.  Must be done prior to calling Terminate, below.
    //
    delete parser;
    delete is;
    // And call the termination method
    XMLPlatformUtils::Terminate();

    free(realSchemaLoc);

    return err_string;

  }

}
  
DOMCountErrorHandler::DOMCountErrorHandler() :

    fSawErrors(false)
{
}

DOMCountErrorHandler::~DOMCountErrorHandler()
{
}

char *getstr(const XMLCh *inp){
  char *ret;
  int i;
  ret = (char *)malloc(1024);
  for(i=0;i<1024 && inp[i];i++)
    ret[i] = inp[i];
  ret[i] = '\0';
  return ret;
}

// ---------------------------------------------------------------------------
//  DOMCountHandlers: Overrides of the SAX ErrorHandler interface
// ---------------------------------------------------------------------------
void DOMCountErrorHandler::error(const SAXParseException& e)
{
    fSawErrors = true;
    const XMLCh *foo = e.getMessage();
    char *sysid;
    char *msg;

    sysid = getstr(e.getSystemId());
    msg = getstr(e.getMessage());

    int newerrlength = snprintf(NULL,0,"%s%s:Line %i, Char %i: %s\n",
				err_string,
				sysid,
				e.getLineNumber(),
				e.getColumnNumber(),
				msg);
    err_string = (char *)realloc(err_string,newerrlength+1);

    sprintf(err_string,"%s%s:Line %i, Char %i: %s\n",
	    err_string,
	    sysid,
	    e.getLineNumber(),
	    e.getColumnNumber(),
	    msg);

    free(sysid);
    free(msg);

    //cerr << "\nError at file " << StrX(e.getSystemId())
    //<< ", line " << e.getLineNumber()
    //<< ", char " << e.getColumnNumber()
    //<< "\n  Message: " << StrX(e.getMessage()) << endl;
}

void DOMCountErrorHandler::fatalError(const SAXParseException& e)
{
    fSawErrors = true;
    const XMLCh *foo = e.getMessage();
    char *sysid;
    char *msg;

    sysid = getstr(e.getSystemId());
    msg = getstr(e.getMessage());
    
    int newerrlength = snprintf(NULL,0,"%s%s:Line %i, Char %i: %s\n",
				err_string,
				sysid,
				e.getLineNumber(),
				e.getColumnNumber(),
				msg);
    err_string = (char *)realloc(err_string,newerrlength+1);
    
    sprintf(err_string,"%s%s:Line %i, Char %i: %s\n",
	    err_string,
	    sysid,
	    e.getLineNumber(),
	    e.getColumnNumber(),
	    msg);

    free(sysid);
    free(msg);
    
    //cerr << "\nFatal Error at file " << StrX(e.getSystemId())
    //<< ", line " << e.getLineNumber()
    //<< ", char " << e.getColumnNumber()
    //<< "\n  Message: " << StrX(e.getMessage()) << endl;
}

void DOMCountErrorHandler::warning(const SAXParseException& e)
{
    const XMLCh *foo = e.getMessage();
    char *sysid;
    char *msg;
    
    sysid = getstr(e.getSystemId());
    msg = getstr(e.getMessage());

    int newerrlength = snprintf(NULL,0,"%s%s:Line %i, Char %i: %s\n",
				err_string,
				sysid,
				e.getLineNumber(),
				e.getColumnNumber(),
				msg);
    err_string = (char *)realloc(err_string,newerrlength+1);
    
    sprintf(err_string,"%s%s:Line %i, Char %i: %s\n",
	    err_string,
	    sysid,
	    e.getLineNumber(),
	    e.getColumnNumber(),
	    msg);

    free(sysid);
    free(msg);

    //cerr << "\nWarning at file " << StrX(e.getSystemId())
    //<< ", line " << e.getLineNumber()
    //<< ", char " << e.getColumnNumber()
    //<< "\n  Message: " << StrX(e.getMessage()) << endl;
}

void DOMCountErrorHandler::resetErrors()
{
    fSawErrors = false;
}
